#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import logging
import os
import sys

from amaranth                 import Cat, DomainRenamer, Elaboratable, Module, ResetSignal, Signal, Record
from amaranth.build           import Attrs, Pins, PinsN, Resource, Subsignal
from amaranth.hdl.rec         import Record

from luna_soc.gateware.vendor.lambdasoc.periph         import Peripheral
from luna_soc.gateware.vendor.lambdasoc.periph.serial  import AsyncSerialPeripheral
from luna_soc.gateware.vendor.amaranth_stdio.serial    import AsyncSerial

from luna                            import configure_default_logging, top_level_cli
from luna.gateware.platform          import NullPin
from luna.gateware.usb.usb2.device   import USBDevice

from luna_soc.gateware.cpu.vexriscv  import VexRiscv
from luna_soc.gateware.lunasoc       import LunaSoC

from luna_soc.gateware.csr           import GpioPeripheral, LedPeripheral
from luna_soc.gateware.csr           import USBDeviceController
from luna_soc.gateware.csr           import SetupFIFOInterface, InFIFOInterface, OutFIFOInterface
from luna_soc.gateware.wishbone      import ECP5ConfigurationFlashInterface, SPIPHYController, SPIFlashPeripheral

from .advertiser import ApolloAdvertiserPeripheral
from .info       import CynthionInformationPeripheral


# - MoondancerSoc ---------------------------------------------------------------

class MoondancerSoc(Elaboratable):
    ADDITIONAL_RESOURCES = [
        # PMOD B: UART
        Resource("uart", 1,
            Subsignal("rx",  Pins("1", conn=("pmod", 1), dir="i")),
            Subsignal("tx",  Pins("2", conn=("pmod", 1), dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # PMOD B: DEBUG
        Resource("debug", 0,
            Subsignal("a",  Pins("3", conn=("pmod", 1), dir="o")),
            Subsignal("b",  Pins("4", conn=("pmod", 1), dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # PMOD B: JTAG
        Resource("jtag", 0,
            Subsignal("tms",  Pins("7",  conn=("pmod", 1), dir="i")),
            Subsignal("tdi",  Pins("8",  conn=("pmod", 1), dir="i")),
            Subsignal("tdo",  Pins("9",  conn=("pmod", 1), dir="o")),
            Subsignal("tck",  Pins("10", conn=("pmod", 1), dir="i")),
            Attrs(IO_TYPE="LVCMOS33")
        ),
    ]

    def __init__(self, clock_frequency=int(60e6), uart_baud_rate=115200):
        # qspi flash configuration
        spi0_flash_size  = 0x00400000
        spi0_flash_addr  = 0x10000000
        spi0_csr_addr    = 0xf0008000
        firmware_start   = 0x000b0000

        # Create our SoC...
        self.soc = LunaSoC(
            cpu=VexRiscv(
                variant="cynthion+jtag",
                reset_addr=spi0_flash_addr + firmware_start,
            ),
            clock_frequency=clock_frequency,
        )

        # ... add some stand-ins for our uart pins ...
        self.uart0_pins = Record([
            ('rx', [('i', 1)]),
            ('tx', [('o', 1)])
        ])
        self.uart1_pins = Record([
            ('rx', [('i', 1)]),
            ('tx', [('o', 1)])
        ])
        self.qspi0_pins = Record([
            ('dq',  [('i', 4), ('o', 4), ('oe', 1)]),
            ('cs',  [('o', 1)]),
        ])

        # ... add a stand-in for the INT pin ...
        self.int_pin = Record([
            ('o', [('o', 1)])
        ])

        # ... add core peripherals: memory, timer, uart ...
        self.soc.add_core_peripherals(
            uart_pins=self.uart0_pins,
            uart_baud_rate=uart_baud_rate,
            internal_sram_size=65536,
            internal_sram_addr=0x40000000,
        )

        # ... add a spi flash peripheral ...
        self.spi0_bus        = ECP5ConfigurationFlashInterface(bus=self.qspi0_pins)
        self.spi0_phy        = SPIPHYController(pads=self.spi0_bus, domain="usb", divisor=0)
        self.spi0 = SPIFlashPeripheral(
            self.spi0_phy,
            with_controller=True,
            controller_name="spi0",
            with_mmap=True,
            mmap_size=spi0_flash_size,
            mmap_name="spiflash",
            domain="usb",
        )
        self.soc.add_peripheral(self.spi0, addr=spi0_flash_addr)
        self.soc.add_peripheral(
            self.spi0.spi_controller,
            addr=spi0_csr_addr,
            as_submodule=False,
        )

        # ... add our LED peripheral, for simple output ...
        self.leds = LedPeripheral()
        self.soc.add_peripheral(self.leds, addr=0xf0001000)

        # ... add two gpio peripherals for our PMOD connectors ...
        self.gpioa = GpioPeripheral(width=8)
        self.gpiob = GpioPeripheral(width=8)
        self.soc.add_peripheral(self.gpioa, addr=0xf0002000)
        self.soc.add_peripheral(self.gpiob, addr=0xf0002100)

        # ... and the core USB controllers and eptri peripherals ...
        self.usb0 = USBDeviceController()
        self.usb0_ep_control = SetupFIFOInterface()
        self.usb0_ep_in = InFIFOInterface()
        self.usb0_ep_out = OutFIFOInterface()
        self.soc.add_peripheral(self.usb0, addr=0xf0003000)
        self.soc.add_peripheral(self.usb0_ep_control, as_submodule=False)
        self.soc.add_peripheral(self.usb0_ep_in, as_submodule=False)
        self.soc.add_peripheral(self.usb0_ep_out, as_submodule=False)

        self.usb1 = USBDeviceController()
        self.usb1_ep_control = SetupFIFOInterface()
        self.usb1_ep_in = InFIFOInterface()
        self.usb1_ep_out = OutFIFOInterface()
        self.soc.add_peripheral(self.usb1, addr=0xf0004000)
        self.soc.add_peripheral(self.usb1_ep_control, as_submodule=False)
        self.soc.add_peripheral(self.usb1_ep_in, as_submodule=False)
        self.soc.add_peripheral(self.usb1_ep_out, as_submodule=False)

        self.usb2 = USBDeviceController()
        self.usb2_ep_control = SetupFIFOInterface()
        self.usb2_ep_in = InFIFOInterface()
        self.usb2_ep_out = OutFIFOInterface()
        self.soc.add_peripheral(self.usb2, addr=0xf0005000)
        self.soc.add_peripheral(self.usb2_ep_control, as_submodule=False)
        self.soc.add_peripheral(self.usb2_ep_in, as_submodule=False)
        self.soc.add_peripheral(self.usb2_ep_out, as_submodule=False)

        # ... add a second uart peripheral ...
        self.uart1 = AsyncSerialPeripheral(core=AsyncSerial(
            data_bits = 8,
            divisor   = int(clock_frequency // uart_baud_rate),
            pins      = self.uart1_pins,
        ))
        self.soc.add_peripheral(self.uart1, addr=0xf0006000)

        # ... add an ApolloAdvertiser peripheral ...
        self.advertiser = ApolloAdvertiserPeripheral(pad=self.int_pin, clk_freq_hz=clock_frequency)
        self.soc.add_peripheral(self.advertiser, addr=0xf0007000)

        # ... add a CynthionInformation peripheral ...
        self.info = CynthionInformationPeripheral()
        self.soc.add_peripheral(self.info, addr=0xf0007100)

    def elaborate(self, platform):
        m = Module()

        # add additional resources (only supported on platforms > r0.4)
        if platform.version not in [(0, 1), (0, 2), (0, 3), (0, 4)]:
            platform.add_resources(self.ADDITIONAL_RESOURCES)

        # generate our domain clocks/resets
        m.submodules.car = platform.clock_domain_generator()

        # add SoC to design and clock it off the 60 MHz "usb" domain
        # because VexriscV synthesis tops out at ~77 MHz
        m.submodules.soc = DomainRenamer({"sync": "usb"})(self.soc)

        # wire up the cpu external reset signal
        try:
            user1_io = platform.request("button_user")
            m.d.comb += self.soc.cpu.ext_reset.eq(user1_io.i)
        except:
            logging.warning("Platform does not support a user button for cpu reset")

        # connect QSPI0 to Cynthion's qspi flash port
        qspi0_io = platform.request("qspi_flash", 0)
        m.d.comb += [
            qspi0_io.dq.oe.eq(self.qspi0_pins.dq.oe),
            qspi0_io.dq.o.eq(self.qspi0_pins.dq.o),
            qspi0_io.cs.o.eq(self.qspi0_pins.cs.o),
            self.qspi0_pins.dq.i.eq(qspi0_io.dq.i),
        ]
        m.submodules += [self.spi0_bus, self.spi0_phy]

        # connect GPIOA to Cynthion's PMOD A port
        try:
            pmoda_io = platform.request("user_pmod", 0)
            m.d.comb += [
                self.gpioa.pins.connect(pmoda_io),
            ]
        except:
            logging.warning("Platform does not support a user pmod port for gpio")

        # connect UART0 to Cynthion's SAMD11 uart
        uart0_io = platform.request("uart", 0)
        m.d.comb += [
            uart0_io.tx.o.eq(self.uart0_pins.tx),
            self.uart0_pins.rx.eq(uart0_io.rx)
        ]
        if hasattr(uart0_io.tx, 'oe'):
            m.d.comb += uart0_io.tx.oe.eq(~self.soc.uart._phy.tx.rdy),

        # connect UART1 to Cynthion's PMOD B port
        try:
            uart1_io = platform.request("uart", 1)
            m.d.comb += [
                uart1_io.tx.o.eq(self.uart1_pins.tx),
                self.uart1_pins.rx.eq(uart1_io.rx)
            ]
        except:
            logging.warning("Platform does not support a user pmod port for a second uart")

        # connect INT pin to ApolloAdvertiser
        try:
            int_io = platform.request("int")
            m.d.comb += [
                int_io.o.eq(self.int_pin)
            ]
        except:
            logging.warning("Platform does not support ApolloAdvertiserPeripheral")

        # connect JTAG0 to Cynthion's PMOD B port
        try:
            jtag0_io = platform.request("jtag", 0)
            m.d.comb += [
                self.soc.cpu.jtag_tms  .eq(jtag0_io.tms.i),
                self.soc.cpu.jtag_tdi  .eq(jtag0_io.tdi.i),
                jtag0_io.tdo.o         .eq(self.soc.cpu.jtag_tdo),
                self.soc.cpu.jtag_tck  .eq(jtag0_io.tck.i),
                self.soc.cpu.dbg_reset .eq(self.soc.cpu.ext_reset),
            ]
        except:
            logging.warning("Platform does not support a user pmod port for jtag")

        # create our USB devices, connect device controllers and add eptri endpoint handlers

        # target_phy
        ulpi0 = platform.request("target_phy")
        usb0_device = USBDevice(bus=ulpi0)
        usb0_device.add_endpoint(self.usb0_ep_control)
        usb0_device.add_endpoint(self.usb0_ep_in)
        usb0_device.add_endpoint(self.usb0_ep_out)
        m.d.comb += self.usb0.attach(usb0_device)
        m.submodules.usb0_device = usb0_device

        # aux_phy
        try:
            ulpi1 = platform.request("aux_phy")
        except:
            ulpi1 = platform.request("host_phy")
        usb1_device = USBDevice(bus=ulpi1)
        usb1_device.add_endpoint(self.usb1_ep_control)
        usb1_device.add_endpoint(self.usb1_ep_in)
        usb1_device.add_endpoint(self.usb1_ep_out)
        m.d.comb += self.usb1.attach(usb1_device)
        m.submodules.usb1_device = usb1_device

        # control_phy
        try:
            ulpi2 = platform.request("control_phy")
        except:
            ulpi2 = platform.request("sideband_phy")
        usb2_device = USBDevice(bus=ulpi2)
        usb2_device.add_endpoint(self.usb2_ep_control)
        usb2_device.add_endpoint(self.usb2_ep_in)
        usb2_device.add_endpoint(self.usb2_ep_out)
        m.d.comb += self.usb2.attach(usb2_device)
        m.submodules.usb2_device = usb2_device

        return m



# - main ----------------------------------------------------------------------

if __name__ == "__main__":
    from luna                    import configure_default_logging
    from luna.gateware.platform  import get_appropriate_platform
    from luna_soc                import top_level_cli

    # configure logging
    configure_default_logging()
    logging.getLogger().setLevel(logging.DEBUG)

    # select platform
    _platform = get_appropriate_platform()
    if _platform is None:
        logging.error("Failed to identify a supported platform")
        sys.exit(1)

    # configure clock frequency
    _clock_frequency = int(_platform.DEFAULT_CLOCK_FREQUENCIES_MHZ["usb"] * 1e6)
    logging.info(f"Building for {_platform} with clock frequency: {_clock_frequency}")

    # create design
    _design = MoondancerSoc(clock_frequency=_clock_frequency)

    # invoke cli
    _overrides = {
        "debug_verilog": False,
        "verbose": False,
    }
    top_level_cli(_design, **_overrides)



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Support functions for Cynthion CLI. """

import argparse
import os
import logging
import platform
import re
import sys
import textwrap
import usb

try:
    from importlib_resources import files # <= 3.8
except:
    from importlib.resources import files # >= 3.9

from cynthion                  import shared
from fwup.dfu                  import DFUTarget
from tqdm                      import tqdm


SOC_FIRMWARE_FLASHADDR = 0x000b0000

_MSG_EDITABLE_INSTALL = """
If you have installed the 'cynthion' Python package from source please
run the following command in the package directory:

    make assets
"""
_MSG_UNSUPPORTED_FIRMWARE = """
There is no prebuilt Apollo firmware available for Cynthion r{}.{}.

Please see the developer documentation for instructions on how to build
and flash your own:

    https://cynthion.readthedocs.io/en/latest/hardware/bringup_guide.html
"""

_MSG_SOURCE_INSTALL = """
If you have installed the 'cynthion' Python package from source please
run the following commands in the package directory:

    make assets
    pip install --upgrade .
"""

def _find_assets_path():
    package_path = files(__package__.split('.')[0])
    if _is_editable_install():
        assets = os.path.join(package_path, "../assets")
    else:
        assets = os.path.join(package_path, "assets")

    return os.path.normpath(os.path.join(package_path, assets))


def _is_editable_install():
    package_name = __package__.split('.')[0]
    package_path = files(package_name)

    # were we installed with `pip -e .` ?
    if os.path.basename(package_path) == "src":
        return True


def _is_source_install():
    package_name = __package__.split('.')[0]
    package_path = files(package_name)

    # get .dist-info path
    dist_info_path = None
    site_packages_path = os.path.dirname(package_path)
    for item in os.listdir(site_packages_path):
        if item.endswith(".dist-info") and item.startswith(package_name.replace('-', '_')):
            dist_info_path = os.path.join(site_packages_path, item)
            break

    # pypi installs will always have a .dist-info/ directory
    if not dist_info_path:
        return True

    # .dist-info/direct_url.json only exists on non-pypi installs (PEP 610)
    return os.path.isfile(os.path.join(dist_info_path, "direct_url.json"))


def find_cynthion_asset(filename):
    """Returns the path to the requested asset filename"""

    asset_path = os.path.join(_find_assets_path(), filename)
    if os.path.isfile(asset_path):
        return asset_path
    else:
        logging.error(f"The Cynthion '{filename}' asset could not be located.")
        if _is_editable_install():
            logging.error(_MSG_EDITABLE_INSTALL)
        elif _is_source_install():
            logging.error(_MSG_SOURCE_INSTALL)
        sys.exit(1)


def find_cynthion_bitstream(device, filename):
    """Returns the path to the requested bitstream for the appropriate platform"""

    platform = _get_appropriate_platform_name(device)

    bitstream_path = os.path.join(_find_assets_path(), platform, filename)
    if os.path.isfile(bitstream_path):
        return bitstream_path
    else:
        logging.error(f"The Cynthion '{filename}' bitstream could not be located.")
        if _is_editable_install():
            logging.error(_MSG_EDITABLE_INSTALL)
        elif _is_source_install():
            logging.error(_MSG_SOURCE_INSTALL)
        sys.exit(1)


def flash_bitstream(device, filename):
    """Flashes the given filename to the FPGA configuration flash"""
    with open(filename, "rb") as f:
        bitstream = f.read()

    logging.info(f"Updating FPGA configuration flash with {len(bitstream)} bytes...")

    device.force_fpga_offline()
    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        programmer.flash(bitstream, offset=0)
    logging.info("Operation complete!")

    # reset device
    device.soft_reset()

    # let the gateware take over in devices with a shared usb port
    device.allow_fpga_takeover_usb()


def flash_mcu_firmware(device, filename):
    """Flashes the given filename to the device's Microcontroller"""
    with open(filename, 'rb') as f:
        firmware = f.read()

    # Check for unsupported hardware
    major, minor = device.detect_connected_version()
    if major == 0 and minor < 6:
        logging.error(_MSG_UNSUPPORTED_FIRMWARE.format(major, minor))
        sys.exit(1)

    # Release Apollo debugger
    device.close()

    # create a DFU programmer instance
    board = DFUTarget(idVendor=shared.usb.bVendorId.apollo, idProduct=shared.usb.bProductId.apollo)

    logging.info(f"Updating device firmware with {len(firmware)} bytes...")

    # program firmware binary
    size_to_program = board.size_to_program(firmware)
    with tqdm(total=size_to_program, ncols=80, unit='B', leave=False) as progress:
        board.program(firmware, status_callback=lambda written, _: progress.update(written))

    logging.info("Operation complete!")

    # restart firmware
    board.run_user_program()


def flash_soc_firmware(device, filename):
    """Flashes the given filename to the section of configuration flash reserved for the SoC"""
    with open(filename, "rb") as f:
        firmware = f.read()

    logging.info(f"Updating SoC firmware flash with {len(firmware)} bytes...")

    device.force_fpga_offline()
    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        programmer.flash(firmware, offset=SOC_FIRMWARE_FLASHADDR)
    logging.info("Operation complete!")


def run_bitstream(device, filename):
    with open(filename, "rb") as f:
        bitstream = f.read()

    logging.info(f"Uploading target bitstream to FPGA with {len(bitstream)} bytes...")

    with device.jtag as jtag:
        programmer = device.create_jtag_programmer(jtag)
        programmer.configure(bitstream)
    logging.info("Operation complete!")

    # let the gateware take over in devices with a shared usb port
    device.allow_fpga_takeover_usb()


def _get_appropriate_platform_name(device):
    """Returns the LUNA platform name for the connected Cynthion"""
    # Retrieve the version of the attached device.
    major, minor = device.detect_connected_version()

    return f"CynthionPlatformRev{major}D{minor}"


class HelpFormatter(argparse.HelpFormatter):
    def __init__(self, prog):
        if "COLUMNS" in os.environ:
            columns = int(os.environ["COLUMNS"])
        else:
            try:
                columns, _ = os.get_terminal_size(sys.stderr.fileno())
            except OSError:
                columns = 80
        super().__init__(prog, width=columns, max_help_position=30)

    def _fill_text(self, text, width, indent):
        def filler(match):
            text = match[0]
            if text.startswith("::"):
                return text[2:]

            list_match = re.match(r"(\s*)(\*.+)", text, flags=re.S)
            if list_match:
                text = re.sub(r"(\S)\s+(\S)", r"\1 \2", list_match[2])
                text = textwrap.fill(text, width,
                                     initial_indent=indent + "  ",
                                     subsequent_indent=indent + "    ")
            else:
                text = textwrap.fill(text, width,
                                     initial_indent=indent,
                                     subsequent_indent=indent)

            text = text + (match[2] or "")
            text = re.sub(r"(\w-) (\w)", r"\1\2", text)
            return text

        text = textwrap.dedent(text).strip()
        return re.sub(r"((?!\n\n)(?!\n\s+(?:\*|\$|\d+\.)).)+(\n*)?", filler, text, flags=re.S)



#!/usr/bin/env python3
#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from . import registers
from . import gateware
from . import host
